# NOTE: since this is imported in models.py, ensure it doesn't use functionality that
# requires django to be fully initialized.
from typing import Optional

from .conf import settings
from .tasks import reset_config


def schedule_config_reset(reset_after: Optional[int]):
    reset_after = reset_after or settings.LOG_OUTGOING_REQUESTS_RESET_DB_SAVE_AFTER
    if not reset_after:
        return

    countdown = reset_after * 60
    reset_config.apply_async(countdown=countdown)



# Generated by Django 3.2.24 on 2024-02-27 09:32

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("log_outgoing_requests", "0005_alter_outgoingrequestslog_url"),
    ]

    operations = [
        migrations.AddField(
            model_name="outgoingrequestslogconfig",
            name="reset_db_save_after",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="If configured, after the config has been updated, reset the database logging after the specified number of minutes. Note: this overrides the LOG_OUTGOING_REQUESTS_RESET_DB_SAVE_AFTER environment variable. Additionally, depending on the broker that is used, if this duration is too long the key for the reset task might have expired before that time. So make sure not to set too large a value for the reset.",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Reset saving logs in database after",
            ),
        ),
    ]



import logging

from .compat import shared_task
from .constants import SaveLogsChoice

logger = logging.getLogger(__name__)


@shared_task
def prune_logs():
    from .models import OutgoingRequestsLog

    num_deleted = OutgoingRequestsLog.objects.prune()
    logger.info("Deleted %d outgoing request log(s)", num_deleted)
    return num_deleted


@shared_task
def reset_config():
    from .models import OutgoingRequestsLogConfig

    config = OutgoingRequestsLogConfig.get_solo()
    assert isinstance(config, OutgoingRequestsLogConfig)

    # nothing to do
    if config.save_to_db == SaveLogsChoice.use_default:
        return

    config.save_to_db = SaveLogsChoice.use_default
    config.save(update_fields=["save_to_db"])



from urllib.parse import urlparse

from django import forms
from django.contrib import admin
from django.utils.html import mark_safe
from django.utils.translation import gettext as _

from solo.admin import SingletonModelAdmin

from .conf import settings
from .models import OutgoingRequestsLog, OutgoingRequestsLogConfig

try:
    import celery
except ImportError:
    celery = None


@admin.register(OutgoingRequestsLog)
class OutgoingRequestsLogAdmin(admin.ModelAdmin):
    list_display = (
        "hostname",
        "truncated_url",
        "params",
        "status_code",
        "method",
        "response_ms",
        "timestamp",
        "response_content_length",
    )
    list_filter = ("method", "timestamp", "status_code", "hostname")
    search_fields = ("url", "params", "hostname")
    date_hierarchy = "timestamp"
    show_full_result_count = False
    readonly_fields = (
        "url",
        "timestamp",
        "method",
        "query_params",
        "prettify_body_response",
        "params",
        "req_headers",
        "req_content_type",
        "request_body",
        "status_code",
        "response_ms",
        "res_headers",
        "res_content_type",
        "response_content_length",
        "response_body",
        "trace",
    )

    def get_fieldsets(self, request, obj=None):
        fieldsets = [
            (
                _("Request"),
                {
                    "fields": (
                        "method",
                        "url",
                        "timestamp",
                        "query_params",
                        "params",
                        "req_headers",
                        "req_content_type",
                        "req_body_encoding",
                        "request_body",
                    )
                },
            ),
            (
                _("Response"),
                {
                    "fields": (
                        "status_code",
                        "response_ms",
                        "res_headers",
                        "res_content_type",
                        "res_body_encoding",
                        "response_content_length",
                        "response_body",
                    )
                },
            ),
            (_("Extra"), {"fields": ("trace",)}),
        ]

        if obj and obj.supports_xml_or_json:
            fieldsets[1][1]["fields"] += ("prettify_body_response",)

        return fieldsets

    class Media:
        css = {
            "all": ("log_outgoing_requests/css/admin.css",),
        }
        js = ("log_outgoing_requests/js/admin.js",)

    def has_add_permission(self, request):
        return False

    @admin.display(description=_("Query parameters"))
    def query_params(self, obj):
        return obj.query_params

    @admin.display(description=_("Request body"))
    def request_body(self, obj) -> str:
        return obj.request_body_decoded or "-"

    @admin.display(description=_("Response body"))
    def response_body(self, obj) -> str:
        return obj.response_body_decoded or "-"

    def prettify_body_response(self, obj):
        body_response = ""
        if obj.supports_xml_or_json:
            body_response = mark_safe(
                '<a href="#" class="prettify-toggle-link">Prettify</a><br>\n'
                '<textarea readonly class="prettify-output" style="display:none;"\n'
                f" rows='15' cols='60' content-type='{obj.res_content_type}'>\n"
                f"{obj.response_body_decoded}</textarea>"
            )
        return body_response

    def truncated_url(self, obj):
        parsed_url = urlparse(obj.url)
        path = parsed_url.path
        max_length = 200
        path_length = len(path)

        if path_length <= max_length:
            return path

        half_length = (max_length - 3) // 2
        left_half = path[:half_length]
        right_half = path[-half_length:]
        return left_half + " \u2026 " + right_half


class ConfigAdminForm(forms.ModelForm):
    class Meta:
        model = OutgoingRequestsLogConfig
        fields = "__all__"
        help_texts = {
            "save_to_db": _(
                "Whether request logs should be saved to the database (default: {default})."
            ).format(default=settings.LOG_OUTGOING_REQUESTS_DB_SAVE),
            "save_body": _(
                "Whether the body of the request and response should be logged (default: "
                "{default})."
            ).format(default=settings.LOG_OUTGOING_REQUESTS_DB_SAVE_BODY),
        }


@admin.register(OutgoingRequestsLogConfig)
class OutgoingRequestsLogConfigAdmin(SingletonModelAdmin):
    form = ConfigAdminForm

    def get_fields(self, request, obj=None, *args, **kwargs):
        fields = super().get_fields(request, obj=obj, *args, **kwargs)
        if celery is None and (obj and not obj.reset_db_save_after):
            fields.remove("reset_db_save_after")
        return fields



# NOTE: Avoid import Django specifics at the module level to prevent circular imports.
# The handler is loaded eagerly at django startup when configuring settings.
import logging
import time
from contextlib import contextmanager
from datetime import timedelta
from typing import Optional, cast
from urllib.parse import urlparse

from django.utils import timezone

from requests import PreparedRequest, RequestException, Response

from .compat import format_exception
from .typing import (
    AnyLogRecord,
    ErrorRequestLogRecord,
    RequestLogRecord,
    is_request_log_record,
)

logger = logging.getLogger(__name__)


@contextmanager
def supress_errors():
    try:
        yield
    except Exception as exc:
        logger.error("Could not persist log record to DB", exc_info=exc)


class DatabaseOutgoingRequestsHandler(logging.Handler):
    """
    Save the log record to the database if conditions are met.

    The handler checks if saving to the database is desired. If not, nothing happens.
    Next, request and response body are each checked if:

        * saving to database is desired
        * the content type is appropriate
        * the size of the body does not exceed the configured treshold

    If any of the conditions don't match, then the body is omitted.
    """

    @supress_errors()
    def emit(self, record: AnyLogRecord):
        from .models import OutgoingRequestsLog, OutgoingRequestsLogConfig
        from .utils import process_body

        # skip requests not coming from the library requests
        if not record or not is_request_log_record(record):
            return

        config = OutgoingRequestsLogConfig.get_solo()
        assert isinstance(config, OutgoingRequestsLogConfig)
        if not config.save_logs_enabled:
            return

        # Python 3.10 TypeGuard can be useful here
        record = cast(RequestLogRecord, record)

        # check if we're dealing with success or error state
        exception = record.exc_info[1] if record.exc_info else None
        if (response := getattr(record, "res", None)) is not None:
            # we have a response - this is the 'happy' flow (connectivity is okay)
            record = cast(RequestLogRecord, record)
            request = record.req
        elif isinstance(exception, RequestException):
            record = cast(ErrorRequestLogRecord, record)
            # we have an requests-specific exception
            request: Optional[PreparedRequest] = exception.request
            response: Optional[Response] = exception.response  # likely None
        else:  # pragma: no cover
            logger.debug("Received log record that cannot be handled %r", record)
            return

        scrubbed_req_headers = request.headers.copy() if request else {}
        if "Authorization" in scrubbed_req_headers:
            scrubbed_req_headers["Authorization"] = "***hidden***"

        parsed_url = urlparse(request.url) if request else None

        # ensure we have a timezone aware timestamp. time.time() is platform dependent
        # about being UTC or a local time. A robust way is checking how many seconds ago
        # this record was created, and subtracting that from the current tz aware time.
        time_delta_logged_seconds = time.time() - record.created
        timestamp = timezone.now() - timedelta(seconds=time_delta_logged_seconds)

        kwargs = {
            "url": request.url if request else "(unknown)",
            "hostname": parsed_url.netloc if parsed_url else "(unknown)",
            "params": parsed_url.params if parsed_url else "(unknown)",
            "status_code": response.status_code if response is not None else None,
            "method": request.method if request else "(unknown)",
            "timestamp": timestamp,
            "response_ms": (
                int(response.elapsed.total_seconds() * 1000)
                if response is not None
                else 0
            ),
            "req_headers": self.format_headers(scrubbed_req_headers),
            "res_headers": self.format_headers(
                response.headers if response is not None else {}
            ),
            "trace": "\n".join(format_exception(exception)) if exception else "",
        }

        if config.save_body_enabled:
            # check request
            if (
                request
                and (
                    processed_request_body := process_body(request, config)
                ).allow_saving_to_db
            ):
                kwargs.update(
                    {
                        "req_content_type": processed_request_body.content_type,
                        "req_body": processed_request_body.content,
                        "req_body_encoding": processed_request_body.encoding,
                    }
                )

            # check response
            if (
                response is not None
                and (
                    processed_response_body := process_body(response, config)
                ).allow_saving_to_db
            ):
                kwargs.update(
                    {
                        "res_content_type": processed_response_body.content_type,
                        "res_body": processed_response_body.content,
                        "res_body_encoding": processed_response_body.encoding,
                    }
                )

        OutgoingRequestsLog.objects.create(**kwargs)

    def format_headers(self, headers):
        return "\n".join(f"{k}: {v}" for k, v in headers.items())



import logging
from datetime import timedelta
from typing import Union
from urllib.parse import urlparse

from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from solo.models import SingletonModel  # type: ignore

from .conf import settings
from .config_reset import schedule_config_reset
from .constants import SaveLogsChoice

logger = logging.getLogger(__name__)


class OutgoingRequestsLogQueryset(models.QuerySet):
    def prune(self) -> int:
        max_age = settings.LOG_OUTGOING_REQUESTS_MAX_AGE
        if max_age is None:
            return 0

        now = timezone.now()
        num_deleted, _ = self.filter(timestamp__lt=now - timedelta(max_age)).delete()
        return num_deleted


class OutgoingRequestsLog(models.Model):
    url = models.TextField(
        verbose_name=_("URL"),
        help_text=_("The url of the outgoing request."),
    )

    # hostname is added so we can filter on it in the admin page
    hostname = models.CharField(
        verbose_name=_("Hostname"),
        max_length=255,
        help_text=_("The netloc/hostname part of the url."),
    )
    params = models.TextField(
        verbose_name=_("Parameters"),
        blank=True,
        help_text=_("The parameters (if they exist)."),
    )
    status_code = models.PositiveIntegerField(
        verbose_name=_("Status code"),
        null=True,
        blank=True,
        help_text=_("The status code of the response."),
    )
    method = models.CharField(
        verbose_name=_("Method"),
        max_length=10,
        blank=True,
        help_text=_("The type of request method."),
    )

    # Request content
    req_content_type = models.CharField(
        verbose_name=_("Request content type"),
        max_length=50,
        help_text=_("The content type of the request."),
    )
    req_headers = models.TextField(
        verbose_name=_("Request headers"), help_text=_("The request headers.")
    )
    req_body_encoding = models.CharField(
        _("Request encoding"),
        max_length=24,
        blank=True,
        help_text=_(
            "The encoding is either extracted from the Content-Type header or, if "
            "absent, taken from the default encoding configured for the content type. "
            "If the decoded request content is not displaying correctly, you may try "
            "changing the encoding value here."
        ),
    )
    req_body = models.BinaryField(
        verbose_name=_("Request body"), default=b"", help_text=_("The request body.")
    )

    # Response content
    res_content_type = models.CharField(
        verbose_name=_("Response content type"),
        max_length=50,
        help_text=_("The content type of the response."),
    )
    res_headers = models.TextField(
        verbose_name=_("Response headers"), help_text=_("The response headers.")
    )
    res_body = models.BinaryField(
        verbose_name=_("Response body"), default=b"", help_text=_("The response body.")
    )
    res_body_encoding = models.CharField(
        _("Response encoding"),
        max_length=24,
        blank=True,
        help_text=_(
            "The encoding is either extracted from the Content-Type header or, if "
            "absent, taken from the default encoding configured for the content type. "
            "If the decoded response content is not displaying correctly, you may try "
            "changing the encoding value here."
        ),
    )
    response_ms = models.PositiveIntegerField(
        verbose_name=_("Response in ms"),
        default=0,
        help_text=_("This is the response time in ms."),
    )
    timestamp = models.DateTimeField(
        verbose_name=_("Timestamp"),
        help_text=_("This is the date and time the API call was made."),
    )
    trace = models.TextField(
        verbose_name=_("Trace"),
        help_text=_("Text providing information in case of request failure."),
    )

    objects = OutgoingRequestsLogQueryset.as_manager()

    class Meta:
        verbose_name = _("Outgoing request log")
        verbose_name_plural = _("Outgoing request logs")

    def __str__(self):
        return ("{hostname} at {date}").format(
            hostname=self.hostname, date=self.timestamp
        )

    @cached_property
    def url_parsed(self):
        return urlparse(self.url)

    @cached_property
    def supports_xml_or_json(self):
        return "xml" in self.res_content_type or "json" in self.res_content_type

    @property
    def query_params(self):
        return self.url_parsed.query

    def _decode_body(self, content: Union[bytes, memoryview], encoding: str) -> str:
        """
        Decode body for use in template.

        If the stored encoding is not found (either because it is empty or because of
        spelling errors etc.), we decode "blindly", replacing chars that could not be
        decoded.

        Inspired on :meth:`requests.models.Response.text`, which is Apache 2.0 licensed.
        """
        try:
            return str(content, encoding, errors="replace")
        except LookupError:
            # A LookupError is raised if the encoding was not found which could
            # indicate a misspelling or similar mistake.
            return str(content, errors="replace")

    @cached_property
    def request_body_decoded(self) -> str:
        """
        Decoded request body for use in template.
        """
        return self._decode_body(self.req_body, self.req_body_encoding)

    request_body_decoded.short_description = _("Request body")  # type: ignore

    @cached_property
    def response_body_decoded(self) -> str:
        """
        Decoded response body for use in template.
        """
        return self._decode_body(self.res_body, self.res_body_encoding)

    response_body_decoded.short_description = _("Response body")  # type: ignore

    @cached_property
    def response_content_length(self) -> int:
        """
        Get Response content length by reading `len(response_body_decoded)`.
        """
        return len(self.response_body_decoded)

    response_content_length.short_description = _("Response content length")  # type: ignore


def get_default_max_content_length():
    """
    Get default value for max content length from settings.
    """
    return settings.LOG_OUTGOING_REQUESTS_MAX_CONTENT_LENGTH


class OutgoingRequestsLogConfig(SingletonModel):
    """Configuration options for request logging."""

    save_to_db = models.CharField(
        _("Save logs to database"),
        max_length=11,
        choices=SaveLogsChoice.choices,
        default=SaveLogsChoice.use_default,
    )
    save_body = models.CharField(
        _("Save request + response body"),
        max_length=11,
        choices=SaveLogsChoice.choices,
        default=SaveLogsChoice.use_default,
    )
    max_content_length = models.IntegerField(
        _("Maximal content size"),
        validators=[MinValueValidator(0)],
        default=get_default_max_content_length,
        help_text=_(
            "The maximal size of the request/response content (in bytes). "
            "If 'Require content length' is not checked, this setting has no effect."
        ),
    )
    reset_db_save_after = models.PositiveIntegerField(
        _("Reset saving logs in database after"),
        null=True,
        blank=True,
        validators=[MinValueValidator(1)],
        help_text=_(
            "If configured, after the config has been updated, reset the database logging "
            "after the specified number of minutes. Note: this overrides the "
            "LOG_OUTGOING_REQUESTS_RESET_DB_SAVE_AFTER environment variable. Additionally, "
            "depending on the broker that is used, if this duration is too long "
            "the key for the reset task might have expired before that time. So make sure not to "
            "set too large a value for the reset."
        ),
    )

    class Meta:
        verbose_name = _("Outgoing request log configuration")

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        schedule_config_reset(self.reset_db_save_after)

    @property
    def save_logs_enabled(self):
        """
        Use configuration option or settings to determine if logs should be saved.
        """
        if self.save_to_db == SaveLogsChoice.use_default:
            return settings.LOG_OUTGOING_REQUESTS_DB_SAVE
        return self.save_to_db == SaveLogsChoice.yes

    @property
    def save_body_enabled(self):
        """
        Use configuration option or settings to determine if log bodies should be saved.
        """
        if self.save_body == SaveLogsChoice.use_default:
            return settings.LOG_OUTGOING_REQUESTS_DB_SAVE_BODY
        return self.save_body == SaveLogsChoice.yes



# Generated by Django 4.1 on 2023-06-08 13:52

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("log_outgoing_requests", "0002_outgoingrequestslogconfig_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="outgoingrequestslog",
            options={
                "verbose_name": "Outgoing request log",
                "verbose_name_plural": "Outgoing request logs",
            },
        ),
        migrations.AlterModelOptions(
            name="outgoingrequestslogconfig",
            options={"verbose_name": "Outgoing request log configuration"},
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslogconfig",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]
